#!/usr/bin/env python3

from .. import model
from ..model import UNSPECIFIED
from ..getters import Getter, Key, ImplicitKey
from ..pick import ValuesIter, first
from ..meta import NeverAccessedMeta, ExceptionMeta, SetAttrMeta
from ..utils import noop
from ..errors import ApiError, NoValueFound, Log
from math import inf

class param:

    class _State:

        def __init__(self, default):
            self.bound_getters = []
            self.default = default
            self.reset()

        def reset(self):
            self.value = UNSPECIFIED
            self.exception = UNSPECIFIED
            self.meta = NeverAccessedMeta()
            self.dynamic = False
            self.cache_version = -1

    def __init__(
            self,
            *keys,
            cast=noop,
            pick=first,
            default=UNSPECIFIED,
            default_factory=UNSPECIFIED,
            ignore=UNSPECIFIED,
            get=lambda obj, x: x,
            dynamic=False,
    ):
        self._keys = keys
        self._cast = cast
        self._pick = pick
        self._default_factory = _merge_default_args(default, default_factory)
        self._ignore = ignore
        self._get = get
        self._dynamic = dynamic

    def __set_name__(self, cls, name):
        self._name = name

    def __get__(self, obj, cls=None):
        return self._load_value(obj)

    def __set__(self, obj, value):
        if value is self._ignore:
            return

        state = self._load_state(obj)
        state.value = value
        state.exception = UNSPECIFIED
        state.meta = SetAttrMeta()
        state.dynamic = False
        state.cache_version = inf

    def __delete__(self, obj):
        state = self._load_state(obj)
        state.reset()

    def __call__(self, get):
        # Allow the descriptor to be used as a decorator.
        self._get = get
        return self

    def _override(self, args, kwargs, skip=frozenset()):
        # Make sure the override arguments match the constructor:
        import inspect
        sig = inspect.signature(self.__init__)
        sig.bind(*args, **kwargs)

        # Override the attributes referenced by the arguments:
        if args:
            self._keys = args

        if 'default' in kwargs or 'default_factory' in kwargs:
            self._default_factory = _merge_default_args(
                    kwargs.pop('default', UNSPECIFIED),
                    kwargs.pop('default_factory', UNSPECIFIED),
            )

        for key in kwargs.copy():
            if key not in skip:
                setattr(self, f'_{key}', kwargs.pop(key))

    def _load_state(self, obj):
        model.init(obj)
        states = model.get_param_states(obj)

        if self._name not in states:
            default = self._default_factory()
            states[self._name] = self._State(default)

        return states[self._name]

    def _load_value(self, obj):
        state = self._load_state(obj)

        model_version = model.get_cache_version(obj)
        is_cache_stale = (
                state.cache_version < model_version or
                state.dynamic or
                self._dynamic
        )
        if is_cache_stale:
            try:
                state.value, values_iter = self._calc_value(obj)
                state.exception = UNSPECIFIED
                state.meta = values_iter.meta
                state.dynamic = values_iter.dynamic

            # Cache the exception indicating that this parameter is missing, 
            # since that is likely to be raised several times (and unlikely to 
            # terminate the program).
            #
            # Note that other exceptions will not update the cache, and 
            # therefore may need to be calculated on each access.  I could 
            # avoid this by catching all exceptions in this block, but that 
            # would make stack traces more confusing.  I think the approach of 
            # catching only `NoValueFound` strikes a good balance, but I'm open 
            # to revisiting this later.

            except NoValueFound as err:
                state.value = UNSPECIFIED
                state.exception = err
                state.meta = ExceptionMeta(err)

            state.cache_version = model_version

        if state.exception is not UNSPECIFIED:
            raise state.exception

        return self._get(obj, state.value)

    def _load_bound_getters(self, obj):
        state = self._load_state(obj)
        model_version = model.get_cache_version(obj)
        if state.cache_version < model_version:
            state.bound_getters = self._calc_bound_getters(obj)
        return state.bound_getters

    def _load_default(self, obj):
        return self._load_state(obj).default

    def _calc_value(self, obj):
        log = Log()

        # Previously, I used the object's normal repr in this message instead 
        # of explicitly deferring to a generic repr.  However, this led to 
        # infinite recursion in cases where a parameter didn't exist but the 
        # repr function tried to access it.  I tried to fix this by adding a 
        # try/except block, but that ended up triggering a core dump in 
        # python==3.8, see #41.  This was very likely due to a bug in python, 
        # but on the principle that logging code should "do no harm" above 
        # anything else, so I decided to just avoid the problem altogether.
        log += f"getting {self._name!r} parameter for {object.__repr__(obj)}"

        bound_getters = self._load_bound_getters(obj)
        default = self._load_default(obj)
        values = ValuesIter(bound_getters, default, log)
        return self._pick(values), values

    def _calc_bound_getters(self, obj):
        from ..configs.configs import Config
        from inspect import isclass

        keys = [
                Key(x) if isclass(x) and issubclass(x, Config) else x
                for x in self._keys or [self._get_default_key()]
        ]
        wrapped_configs = model.get_wrapped_configs(obj)
        are_getters = [isinstance(x, Getter) for x in keys]

        if all(are_getters):
            getters = keys

        elif any(are_getters):
            err = ApiError(
                    keys=keys,
            )
            err.brief = "can't mix string keys with Key/Method/Func/Value objects"
            err.info += lambda e: '\n'.join((
                    "keys:",
                    *map(repr, e['keys']),
            ))
            raise err

        elif len(keys) == 1:
            getters = [
                    ImplicitKey(wrapped_config, keys[0])
                    for wrapped_config in wrapped_configs
            ]

        elif len(keys) != len(wrapped_configs):
            err = ApiError(
                    configs=[x.config for x in wrapped_configs],
                    keys=keys,
            )
            err.brief = "number of keys must match number of configs"
            err.info += lambda e: '\n'.join((
                    f"configs ({len(e.configs)}):",
                    *map(repr, e.configs),
            ))
            err.blame += lambda e: '\n'.join((
                    f"keys ({len(e['keys'])}):",
                    *map(repr, e['keys']),
            ))
            raise err

        else:
            getters = [
                    ImplicitKey(wrapped_config, key)
                    for key, wrapped_config in zip(keys, wrapped_configs)
            ]

        bound_getters = [
                getter.bind(obj, self)
                for getter in getters
        ]
        return bound_getters

    def _get_default_key(self):
        return self._name

    def _get_default_cast(self):
        return self._cast

    def _get_known_getter_kwargs(self):
        return {'cast'}

def _merge_default_args(instance, factory):
    have_instance = instance is not UNSPECIFIED
    have_factory = factory is not UNSPECIFIED

    if have_instance and have_factory:
        err = ApiError(
                instance=instance,
                factory=factory,
        )
        err.brief = "can't specify 'default' and 'default_factory'"
        err.info += "default: {instance}"
        err.info += "default_factory: {factory}"
        raise err

    if have_factory:
        return factory
    else:
        return lambda: instance




#!/usr/bin/env python3

import pytest
import doctest
import shelldoctest
import os, re
import subprocess

from pathlib import Path
from dataclasses import dataclass, field
from textwrap import dedent, indent
from traceback import format_exception
from typing import List

DOC_DIR = Path(__file__).parents[1] / 'docs'
RST_GLOB = DOC_DIR.glob('**/*.rst')

def parse_doc_examples(path):
    # States:
    # - text
    # - text-block
    # - tab
    # - tab-block

    @dataclass
    class Block:
        name: str
        indent: int
        lines: List[str] = field(default_factory=list)
        lineno: int = 0

    @dataclass
    class Context:
        indent: int
        blocks: List[Block] = field(default_factory=list)

    curr_state = 'text'
    curr_context = Context('')
    curr_block = None

    contexts = [curr_context]

    directive_pattern = r'^(?P<indent>\s*)\.\. (?P<directive>{})::(\s(?P<arg>.*))?'
    tab_pattern = directive_pattern.format('tab')
    block_pattern = directive_pattern.format('code-block')
    indent_pattern = r'^(?P<indent>\s*).*'

    def iter_lines(path):
        with path.open() as f:
            for lineno, line in enumerate(f, 1):
                yield lineno, line

    def parse_line(line, lineno):
        nonlocal curr_block
        nonlocal curr_context

        if curr_state == 'text':

            # Found tab: create a new context
            m = re.match(tab_pattern, line)
            if m:
                curr_context = Context(m['indent'])
                curr_block = Block(m['arg'], m['indent'])
                curr_context.blocks.append(curr_block)
                contexts.append(curr_context)
                return 'tab'

            # Found code block: add to current context
            m = re.match(block_pattern, line)
            if m:
                curr_block = Block(m['arg'] or 'python', m['indent'])
                curr_context.blocks.append(curr_block)
                return 'text-block'

            # Otherwise: stay in 'text' state
            return 'text'

        if curr_state == 'tab':

            # Found code block within tab:
            m = re.match(block_pattern, line)
            if m:
                if len(m['indent']) > len(curr_context.indent):
                    return 'tab-block'

            if not line.strip():
                return curr_state

            raise AssertionError(f"{path}:{lineno}: unexpected content in '.. tab::' directive: {line!r}\nexpected: '.. code-block::'")

        if curr_state == 'text-block':
            assert curr_block

            # Found end of code block: back to text mode
            m = re.match(indent_pattern, line)
            if len(m['indent']) <= len(curr_block.indent):
                return 'text'

            if not curr_block.lines:
                curr_block.lineno = lineno

            curr_block.lines.append(line)
            return 'text-block'

        if curr_state == 'tab-block':
            assert curr_block

            # Found tab: add to current context
            m = re.match(tab_pattern, line)
            if m:
                curr_block = Block(m['arg'], m['indent'])
                curr_context.blocks.append(curr_block)
                return 'tab'

            # Found end of code block: back to text mode
            m = re.match(indent_pattern, line)
            if len(m['indent']) <= len(curr_context.indent):
                return 'text'

            if len(m['indent']) <= len(curr_block.indent):
                raise AssertionError(f"{path}:{lineno}: unexpected content in '.. tab:: directive: {line!r}")

            if not curr_block.lines:
                curr_block.lineno = lineno

            curr_block.lines.append(line)
            return 'tab-block'

    for lineno, line in iter_lines(path):
        curr_state = parse_line(line, lineno)

    return contexts


@pytest.mark.parametrize(
        'rst_path', [
            pytest.param(p, id=f'{p.relative_to(DOC_DIR).stem}')
            for p in RST_GLOB
        ],
)
def test_doc_examples(rst_path, tmp_path):
    contexts = parse_doc_examples(rst_path)
    runner = doctest.DocTestRunner()
    parsers = {
            'python': doctest.DocTestParser(),
            'bash': shelldoctest.ShellDocTestParser(),
    }
    examples = []
    rst_name = str(rst_path.relative_to(DOC_DIR))

    for i, context in enumerate(contexts, 1):
        test_dir = tmp_path / f'context_{i}'
        test_dir.mkdir()

        examples += [
                doctest.Example(f'chdir({str(test_dir)!r})', ''),
        ]

        for block in context.blocks:
            source = dedent(''.join(block.lines))

            if block.name in parsers:
                parser = parsers[block.name]
                examples_ = parser.get_examples(source, rst_name)

                for example in examples_:
                    example.lineno += block.lineno

                examples += examples_

            else:
                file_path = test_dir / block.name
                file_path.write_text(source)

    class PytestRunner(doctest.DocTestRunner):

        def report_failure(self, out, test, example, got):
            pytest.fail(f"doctest gave unexpected output\nexample:\n{indent(example.source, '  ')}\nwanted:\n{indent(example.want, '  ')}\ngot:\n{indent(got, '  ')}\nlocation:\n  {rst_name}:{example.lineno}")

        def report_unexpected_exception(self, out, test, example, exc_info):
            pytest.fail(f"doctest raised unexpected exception\nexample:\n{indent(example.source, '  ')}\nerror:\n{indent(''.join(format_exception(*exc_info)), '  ')}\nlocation:\n  {rst_name}:{example.lineno}")

    class EvalOutputChecker(doctest.OutputChecker):

        def check_output(self, want, got, optionflags):
            if super().check_output(want, got, optionflags):
                return True

            # If the strings don't match, see if they compare equal when 
            # evaluated.  This accounts for things like the non-deterministic 
            # order in which items are stored in sets.
            try:
                return eval(want) == eval(got)
            except Exception:
                return False

    globs = {
            'chdir': os.chdir,
            'shell': shelldoctest.shell,
    }
    tests = doctest.DocTest(examples, globs, rst_name, rst_name, None, None)

    runner = PytestRunner(
            checker=EvalOutputChecker(),
            optionflags=doctest.IGNORE_EXCEPTION_DETAIL,
    )
    runner.run(tests)



#!/usr/bin/env python3

from .errors import *
from collections.abc import Iterable

def noop(x):
    return x

def first_specified(*values, **kwargs):
    unspecified = kwargs.get('sentinel', None)

    for x in values:
        if x is not unspecified:
            return x

    try:
        return kwargs['default']
    except KeyError:
        err = ApiError(
                values=values,
                sentinel=unspecified,
        )
        err.brief = "must specify a value"
        err.blame += lambda e: f"given {len(e['values'])} {e.sentinel} values"
        err.hints += "did you mean to specify a default value?"
        raise err from None

def lookup(obj, key):
    """
    Lookup the given key in the given object.

    Arguments:
        obj: Any object.
        key:
            If callable: The callable will be called with the given object as 
            its only argument.  Whatever it returns will be taken as the value 
            of the key.

            If non-string iterable: The iterable will be considered as a 
            sequence of keys to iteratively lookup in the object.  In other 
            words, the return value will be something like 
            ``obj[key[0]][key[1]]...``.

            If anything else: The key will be looked up in the given object 
            like so: ``obj[key]``.
    """
    if callable(key):
        return key(obj)

    if isinstance(key, Iterable) and not isinstance(key, str):
        for subkey in key:
            obj = obj[subkey]
        return obj

    return obj[key]





import sys, os
import byoc
from pathlib import Path

## General

project = 'BYOC'
copyright = '2020, Kale Kundert'
version = byoc.__version__
release = byoc.__version__

master_doc = 'index'
source_suffix = '.rst'
templates_path = ['_templates']
exclude_patterns = ['_build']
default_role = 'any'

## Extensions

extensions = [
        'autoclasstoc',
        'sphinx.ext.autodoc',
        'sphinx.ext.autosummary',
        'sphinx.ext.viewcode',
        'sphinx.ext.intersphinx',
        'sphinx.ext.napoleon',
        'sphinx_rtd_theme',
        'sphinx_inline_tabs',
        'sphinx_copybutton',
]
intersphinx_mapping = { #
        'python': ('https://docs.python.org/3', None),
}
autosummary_generate = True
autodoc_default_options = {
        'exclude-members': '__dict__,__weakref__,__module__',
}
copybutton_prompt_text = r">>> |\.\.\. |\$ "
copybutton_prompt_is_regexp = True

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']
html_css_files = ['css/tweaks.css']
pygments_style = 'sphinx'

def remove_tabs_js(app, exc):
    # This javascript keeps the tabs in sync, e.g. assuming that they represent 
    # languages and that you'd always want to see the same one.  That doesn't 
    # apply at all in my case, and it seems like the easiest work-around is 
    # just to nuke the javascript altogether.
    if app.builder.format == 'html' and not exc:
        tabs_js = Path(app.builder.outdir) / '_static' / 'tabs.js'
        tabs_js.unlink()

def setup(app):
    app.connect('build-finished', remove_tabs_js)



#!/usr/bin/env python3

from .errors import ApiError

CONFIG_ATTR = '__config__'
STATE_ATTR = '__byoc__'

UNSPECIFIED = object()

class State:

    def __init__(self, obj):
        self.wrapped_configs = [
                WrappedConfig(cf(obj))
                for cf in get_config_factories(obj)
        ]
        self.param_states = {}
        self.cache_version = 0
        self.load_callbacks = get_load_callbacks(obj).values()

class WrappedConfig:
    # The purpose of this wrapper is to cache the layers generated by each 
    # config.  This avoids making the Config class itself responsible for 
    # caching, which would be an extra burden for users looking to implement 
    # new Config subclasses.

    def __init__(self, config):
        self.config = config
        self.layers = []
        self.is_loaded = False

    def __repr__(self):
        return f'{self.__class__.__name__}({self.config!r}, is_loaded={self.is_loaded!r})'

    def __iter__(self):
        yield from self.layers

    def load(self):
        # While it's tempting to defer actually loading the config until the 
        # layers are needed (e.g. in `__iter__()`), this doesn't work.  The 
        # problem is that configs are allowed to access previous (but not 
        # upcoming) configs during loading.  If loading is deferred, the object 
        # won't have the appropriate configs in place anymore.
        #
        # It may be possible to defer loading configs that promise not to 
        # access prior configs during loading.  But my initial instinct is that 
        # this would be too likely to cause bugs, and too unlikely to actually 
        # save any time.  See #11 for more discussion.
        self.layers = list(self.config.load())
        self.is_loaded = True

def init(obj):
    if hasattr(obj, STATE_ATTR):
        return False

    setattr(obj, STATE_ATTR, State(obj))

    _load_configs(
            obj,
            predicate=lambda wc: wc.config.autoload,
    )

    return True

def load(obj, config_cls=None):
    init(obj)

    _load_configs(
            obj,
            predicate=lambda wc: (
                not wc.is_loaded and
                _is_selected_by_cls(wc.config, config_cls)
            ),
    )

def reload(obj, config_cls=None):
    if init(obj):
        return

    _load_configs(
            obj,
            predicate=lambda wc: (
                wc.is_loaded and 
                _is_selected_by_cls(wc.config, config_cls)
            ),
    )

def append_config(obj, config_factory):
    append_configs(obj, [config_factory])

def append_configs(obj, config_factories):
    init(obj)
    state = get_state(obj)
    i = len(state.wrapped_configs)
    insert_configs(obj, i, config_factories)

def prepend_config(obj, config_factory):
    prepend_configs(obj, [config_factory])

def prepend_configs(obj, config_factories):
    insert_configs(obj, 0, config_factories)

def insert_config(obj, i, config_factory):
    insert_configs(obj, i, [config_factory])

def insert_configs(obj, i, config_factories):
    init(obj)

    new_wrapped_configs = [
            WrappedConfig(cf(obj))
            for cf in config_factories
    ]

    state = get_state(obj)
    state.wrapped_configs = \
            state.wrapped_configs[:i] + \
            new_wrapped_configs + \
            state.wrapped_configs[i:]

    _load_configs(
            obj,
            predicate=lambda wc: wc in new_wrapped_configs
    )

def share_configs(donor, acceptor):
    init(donor); init(acceptor)

    donor_state = get_state(donor)
    acceptor_state = get_state(acceptor)

    acceptor_state.wrapped_configs.extend(donor_state.wrapped_configs)
    acceptor_state.cache_version += 1

def get_state(obj):
    return getattr(obj, STATE_ATTR)

def get_config_factories(obj):
    try:
        return getattr(obj, CONFIG_ATTR)
    except AttributeError:
        return []

def get_wrapped_configs(obj):
    return get_state(obj).wrapped_configs

def get_cache_version(obj):
    return get_state(obj).cache_version

def get_load_callbacks(obj):
    from .configs.on_load import OnLoad

    hits = {}

    for cls in reversed(obj.__class__.__mro__):
        for k, v in cls.__dict__.items():
            if isinstance(v, OnLoad):
                hits[k] = v

    return hits

def get_param_states(obj):
    return get_state(obj).param_states

def get_meta(obj, param):
    from .meta import NeverAccessedMeta

    try:
        states = get_param_states(obj)
    except AttributeError:
        return NeverAccessedMeta()

    try:
        state = states[param]
    except KeyError:
        return NeverAccessedMeta()

    return state.meta


def _load_configs(obj, predicate):
    state = get_state(obj)
    state.wrapped_configs, wrapped_configs = [], state.wrapped_configs
    state.cache_version += 1
    updated_configs = []

    # Rebuild the `wrapped_configs` list from scratch and iterate through the 
    # configs in reverse order so that each config, when being loaded, can make 
    # use of values loaded by previous configs but not upcoming ones.
    for wrapped_config in reversed(wrapped_configs):
        if predicate(wrapped_config):
            wrapped_config.load()
            updated_configs.append(wrapped_config.config)

        state.wrapped_configs.insert(0, wrapped_config)
        state.cache_version += 1

    for callback in state.load_callbacks:
        if callback.is_relevant(updated_configs):
            callback(obj)

def _is_selected_by_cls(config, config_cls):
    return not config_cls or isinstance(config, config_cls)





#!/usr/bin/env python3

from .. import model
from .param import param, UNSPECIFIED
from ..utils import noop
from ..errors import NoValueFound
from more_itertools import partition, first

class Toggle:

    def __init__(self, value):
        self.value = value

def pick_toggled(values):
    bases, toggles = partition(
            lambda x: isinstance(x, Toggle),
            values,
    )

    toggle = first(toggles, Toggle(False))

    try:
        base = first(bases)
    except ValueError:
        raise NoValueFound("can't find base value to toggle", values.log)

    return toggle.value != base

class toggle_param(param):

    def __init__(
            self,
            *keys,
            cast=noop,
            default=UNSPECIFIED,
            ignore=UNSPECIFIED,
            get=lambda obj, x: x,
            dynamic=False,
    ):
        super().__init__(
            *keys,
            cast=cast,
            pick=pick_toggled,
            default=default,
            ignore=ignore,
            get=get,
            dynamic=dynamic,
        )

    def _calc_bound_getters(self, obj):
        bound_getters = super()._calc_bound_getters(obj)

        for bg in bound_getters:
            if bg.kwargs.get('toggle', False):
                bg.cast_funcs.append(Toggle)

        return bound_getters

    def _get_known_getter_kwargs(self):
        return super()._get_known_getter_kwargs() | {'toggle'}



#!/usr/bin/env python3

from .model import UNSPECIFIED
from .meta import UnknownMeta, DefaultMeta
from .errors import NoValueFound
from typing import Iterable
from funcy import autocurry

class ValuesIter:

    # In the future, I might also teach this class to somehow indicate where 
    # each value was loaded from.

    def __init__(self, getters, default, log):
        self.getters = getters
        self.default = default
        self.log = log

        # Output attributes:
        self.dynamic = False
        self.meta = UnknownMeta()  # meant to be set by picker.

    def __iter__(self):
        yield from (v for v, m in self.with_meta)

    @property
    def with_meta(self):
        have_value = False
        self.dynamic = False

        if not self.getters:
            self.log += "nowhere to look for values"

        for getter in self.getters:
            for value, meta, dynamic in getter.iter_values(self.log):
                have_value = True
                self.dynamic = self.dynamic or dynamic
                yield getter.cast_value(value, meta), meta

        if self.default is not UNSPECIFIED:
            have_value = True
            self.log += lambda: f"got default value: {self.default!r}"
            yield self.default, DefaultMeta()

        if not have_value:
            self.log += "did you mean to provide a default?"

def first(it: Iterable):
    try:
        value, meta = next(iter(it.with_meta))
        it.meta = meta
        return value
    except StopIteration as err:
        raise NoValueFound("can't find value for parameter", it.log) from None

def list(it: Iterable):
    from builtins import list

    items = list(zip(*it.with_meta))

    if items:
        values, metas = map(list, items)
    else:
        values, metas = [], []

    it.meta = metas
    return values

@autocurry
def merge_dicts(it: Iterable, keep_last=False):
    values = {}
    it.meta = {}

    for dict_, meta in it.with_meta:
        for key, value in dict_.items():
            if (key not in values) or keep_last:
                values[key] = value
                it.meta[key] = meta

    return values






#!/usr/bin/env python3

import byoc
import pytest
import parametrize_from_file
import sys, os, shlex, re

from byoc.errors import Log
from pytest_unordered import unordered
from unittest import mock
from pathlib import Path
from param_helpers import *

@pytest.fixture
def tmp_chdir(tmp_path):
    import os
    try:
        cwd = os.getcwd()
        os.chdir(tmp_path)
        yield tmp_path
    finally:
        os.chdir(cwd)


@parametrize_from_file(
        schema=[
            cast(
                factory=with_byoc.exec(get='DummyConfig', defer=True),
                expected=with_py.eval,
            ),
            with_byoc.error_or('expected')
        ],
)
def test_config_init(factory, expected, error):
    class DummyObj:
        pass

    factory = factory()
    obj = DummyObj()

    with error:
        config = factory(obj)

        for attr, value in expected.items():
            assert getattr(config, attr) == value

@parametrize_from_file(
        schema=cast(
            obj=with_byoc.exec(get=get_obj),
            invocations=Schema([{
                'argv': shlex.split,
                'expected': {str: with_py.eval},
            }]),
        ),
)
def test_argparse_docopt_config(monkeypatch, obj, usage, brief, invocations):
    from copy import copy

    for invocation in invocations:
        print(invocation)

        test_obj = copy(obj)
        test_argv = invocation['argv']
        test_expected = invocation['expected']

        # Clear `sys.argv` so that if the command-line is accessed prematurely, 
        # e.g. in `init()` rather than `load()`, an error is raised.  Note that 
        # `sys.argv[0]` needs to be present, because `argparse` checks this 
        # when generating usage text.
        monkeypatch.setattr(sys, 'argv', ['app'])

        # These attributes should be available even before `init()` is called.  
        # Note that accessing these attributes may trigger `init()`, e.g. if 
        # the usage text contains default values based on BYOC-managed 
        # attributes.
        if sys.version_info[:2] >= (3, 10):
            usage = re.sub('(?m)^optional arguments:$', 'options:', usage)

        assert test_obj.usage == usage
        assert test_obj.brief == brief

        # Make sure that calling `init()` (if it wasn't implicitly called 
        # above) doesn't cause the command line to be read.
        byoc.init(test_obj)

        monkeypatch.setattr(sys, 'argv', test_argv)
        byoc.load(test_obj)

        for attr, value in test_expected.items():
            assert getattr(test_obj, attr) == value

@mock.patch.dict(os.environ, {"x": "1"})
def test_environment_config():
    class DummyObj:
        __config__ = [byoc.EnvironmentConfig]
        x = byoc.param()

    obj = DummyObj()
    assert obj.x == "1"

@parametrize_from_file(
        schema=cast(
            obj=with_byoc.exec(get=get_obj),
            slug=with_py.eval,
            author=with_py.eval,
            version=with_py.eval,
            files=Schema(files.schema),
            layers=eval_config_layers,
        ),
        indirect=['files'],
)
def test_appdirs_config(tmp_chdir, monkeypatch, obj, slug, author, version, files, layers):
    import appdirs

    class AppDirs:

        def __init__(self, slug, author, version):
            self.slug = slug
            self.author = author
            self.version = version

            self.user_config_dir = 'user'
            self.site_config_dir = 'site'

    monkeypatch.setattr(appdirs, 'AppDirs', AppDirs)

    assert obj.dirs.slug == slug
    assert obj.dirs.author == author
    assert obj.dirs.version == version
    assert list(obj.config_paths) == unordered([
        Path(x)
        for x in files.manifest.keys()
    ])

    byoc.init(obj)
    assert collect_layers(obj)[0] == layers

@parametrize_from_file(
        schema=[
            cast(
                config=with_byoc.eval(defer=True),
                config_cls=with_byoc.eval
            ),
            with_byoc.error_or('name', 'config_cls'),
        ],
)
def test_appdirs_config_get_name_and_config_cls(config, name, config_cls, error):
    config = config()(Mock())

    with error:
        actual_name, actual_config_cls = config.get_name_and_config_cls()
        assert actual_name == name
        assert actual_config_cls == config_cls

@parametrize_from_file(
        schema=[
            cast(
                obj=with_byoc.exec(get=get_obj),
                files=Schema(files.schema),
                layers=eval_config_layers,
            ),
            defaults(load_status=[]),
        ],
        indirect=['files'],
)
def test_file_config(tmp_chdir, obj, files, layers, load_status):
    byoc.init(obj)
    assert collect_layers(obj)[0] == layers

    log = log_load_status(obj)
    assert_log_matches(log, load_status)

def test_file_config_load_status():

    class DummyObj:
        __config__ = [byoc.YamlConfig]
        x = byoc.param()

    obj = DummyObj()

    with pytest.raises(byoc.NoValueFound) as err:
        obj.x

    assert err.match(r"failed to get path\(s\):")
    assert err.match(r"raised AttributeError: 'DummyObj' object has no attribute 'path'")

@parametrize_from_file
def test_on_load(prepare, load, expected):

    class DummyConfig(byoc.Config):
        def load(self):
            yield byoc.DictLayer(values={})

    class A(DummyConfig):
        pass

    class A1(A):
        pass

    class A2(A):
        pass

    class B(DummyConfig):
        autoload = False

    class B1(B):
        pass

    class B2(B):
        pass

    class DummyObj:
        __config__ = [A1, A2, B1, B2]

        def __init__(self):
            self.calls = set()

        @byoc.on_load
        def on_default(self):
            self.calls.add('default')

        @byoc.on_load(DummyConfig)
        def on_dummy_config(self):
            self.calls.add('DummyConfig')

        @byoc.on_load(A)
        def on_a(self):
            self.calls.add('A')

        @byoc.on_load(A1)
        def on_a1(self):
            self.calls.add('A1')

        @byoc.on_load(A2)
        def on_a2(self):
            self.calls.add('A2')

        @byoc.on_load(B)
        def on_b(self):
            self.calls.add('B')

        @byoc.on_load(B1)
        def on_b1(self):
            self.calls.add('B1')

        @byoc.on_load(B2)
        def on_b2(self):
            self.calls.add('B2')

    obj = DummyObj()

    with_locals = Namespace(with_byoc, locals())
    with_locals.exec(prepare)

    obj.calls = set()

    with_locals.exec(load)
    assert obj.calls == set(expected or [])

def test_on_load_inheritance():

    class DummyConfig(byoc.Config):
        def load(self):
            yield byoc.DictLayer(values={})

    class P:
        __config__ = [DummyConfig]

        def __init__(self):
            self.calls = set()

        @byoc.on_load
        def a(self):
            self.calls.add('P/a')

        @byoc.on_load
        def b(self):
            self.calls.add('P/b')

        @byoc.on_load
        def c(self):
            self.calls.add('P/c')

    class F1(P):

        @byoc.on_load
        def a(self):
            self.calls.add('F1/a')

        @byoc.on_load
        def b(self):
            self.calls.add('F1/b')

    class F2(F1):

        @byoc.on_load
        def a(self):
            self.calls.add('F2/a')

    p = P()
    f1 = F1()
    f2 = F2()

    byoc.init(p)
    byoc.init(f1)
    byoc.init(f2)

    assert p.calls  == { 'P/a',  'P/b', 'P/c'}
    assert f1.calls == {'F1/a', 'F1/b', 'P/c'}
    assert f2.calls == {'F2/a', 'F1/b', 'P/c'}

@parametrize_from_file(
        schema=[
            cast(
                f=with_byoc.exec(get='f'),
                raises=with_py.eval,
                x=with_py.eval,
                expected=with_py.eval,
            ),
            defaults(raises=[]),
            error_or('x', 'expected'),
        ],
)
@pytest.mark.parametrize(
        'factory', [
            pytest.param(
                lambda f, raises: byoc.dict_like(*raises)(f),
                id='decorator',
            ),
            pytest.param(
                lambda f, raises: byoc.dict_like(f, *raises),
                id='constructor',
            ),
        ]
)
def test_dict_like(factory, f, raises, x, expected, error):
    with error:
        g = factory(f, raises)
        assert g[x] == expected

def test_dict_like_repr():
    d = byoc.dict_like(int)
    assert repr(d) == "dict_like(<class 'int'>)"


def log_load_status(obj):
    log = Log()

    for wc in byoc.model.get_wrapped_configs(obj):
        wc.config.load_status(log)

    return log




#!/usr/bin/env python3

import byoc
import parametrize_from_file
from param_helpers import *

@parametrize_from_file(
        schema=cast(
            obj=with_byoc.exec(get=get_obj, defer=True),
            expected=Schema({str: eval_meta}),
        ),
)
def test_meta_view(obj, expected):
    obj = obj()
    for param, meta in expected.items():
        assert getattr(obj.meta, param) == meta

def test_meta_repr():
    from byoc.meta import GetterMeta, LayerMeta
    from unittest.mock import Mock

    meta = GetterMeta(Mock())
    assert repr(meta) == 'GetterMeta()'

    layer = Mock()
    layer.location = '/path/to/file'
    meta = LayerMeta(Mock(), layer)
    assert repr(meta) == "LayerMeta(location='/path/to/file')"



#!/usr/bin/env python3

from byoc.errors import NoValueFound, Log

def test_no_value_found_formatting():
    log = Log()
    log += 'b'
    log += 'c'

    err = NoValueFound('a', log)

    assert str(err) == 'a\n• b\n• c'

def test_log():
    log = Log()
    log += 'a'
    log += lambda: 'b'
    log.info('c')
    log.info(lambda: 'd')

    assert log.message_strs == ['a', 'b', 'c', 'd']
    assert log.format() == str(log) == '• a\n• b\n• c\n• d'



#!/usr/bin/env python3

import textwrap
from tidyexc import Error

class ApiError(Error):
    """
    For errors using the API, e.g. calling function with inconsistent/redundant 
    arguments.

    These errors are always caused by the programmer, and can never be 
    triggered by end-user input.
    """
    pass

class NoValueFound(AttributeError):
    """
    The default exception raised when no value can be found for a parameter.

    BYOC tries to avoid raising or interpreting any exceptions relating to 
    accessing parameter values.  Instead, user-provided callbacks are expected 
    to raise if they notice something wrong.  This puts the user in control of 
    exception handling and error messages, both good things for a 
    general-purpose framework like this.

    `NoValueFound` is a bit of an exception to this philosophy.  It's raised by 
    the default picker (`first`) in the event that no values were found for an 
    parameter.  It's interpreted by some parts of BYOC (specifically the 
    `Method` and `Func` getters) to mean that an attempt to get a value should 
    be silently skipped.  Both of these behaviors can be overridden, but 
    they're useful defaults.
    """

    def __init__(self, message, log):
        super().__init__(message + '\n' + log.format())

class Log:
    """
    A record of the search for a parameter value.

    Messages are added to the log using the ``+=`` operator::

        >>> log = Log()
        >>> log += "hello world!"

    Each message can either be a string or a no-argument callable returning a 
    string.  The callables will be evaluated just before the log is displayed, 
    and can be used to defer the construction of strings that are likely to be 
    expensive and/or unlikely to be useful.  For example, it's common to use 
    callables to log when a value is successfully found, because (i) the log is 
    usually only displayed in cases where no value is found and (ii) formatting 
    arbitrary objects could be expensive.
    """
    
    def __init__(self):
        self._messages = []

    def __str__(self):
        return self.format()

    def __iadd__(self, message):
        """
        Add the given message to the log.
        """
        self._messages.append(message)
        return self

    def info(self, message):
        """
        Add the given message to the log.

        The ``+=`` operator is the preferred way to do this.  Only use this 
        method if the ``+=`` operator is inconvenient for some reason (e.g. its 
        not allowed in lambda functions).
        """
        self += message

    @property
    def message_strs(self):
        return [
                m() if callable(m) else m
                for m in self._messages
        ]

    def format(self):
        out = ''
        bullet = '• '
        indent = ' ' * len(bullet)

        for msg in self.message_strs:
            msg = textwrap.indent(msg, indent)
            msg = bullet + msg[len(bullet):]
            out += msg + '\n'

        return out[:-1]




#!/usr/bin/env python3

from . import model
from .meta import meta_view

class BareMeta(type):
    """
    A metaclass that allows a class to be instantiated either in the usual way, 
    or without calling the constructor.  The latter is useful if the object 
    will be initialized in another way, e.g. `byoc.param()` parameters that 
    read from the command line.
    """

    def from_bare(cls):
        self = super(cls, cls).__new__(cls)
        if hasattr(self, '__bareinit__'):
            self.__bareinit__()
        return self

    def __call__(cls, *args, **kwargs):
        self = cls.from_bare()
        self.__init__(*args, **kwargs)
        return self

class App(metaclass=BareMeta):
    meta = meta_view()

    def __bareinit__(self):
        pass

    @classmethod
    def entry_point(cls):
        app = cls.from_bare()
        app.main()

    def main(self):
        raise NotImplementedError

    def load(self, config_cls=None):
        model.load(self, config_cls)

    def reload(self, config_cls=None):
        model.reload(self, config_cls)





#!/usr/bin/env python3

import byoc
import parametrize_from_file
from param_helpers import *

@parametrize_from_file(
        schema=[
            with_byoc.error_or('expected'),
            with_py.eval,
        ],
)
def test_toggle(layers, expected, error):

    class BaseConfig(byoc.Config):

        def load(self):
            yield byoc.DictLayer(
                    values={'flag': self.value},
                    location=self.location,
            )

    configs = []
    toggles = set()
    keys = []

    for i, layer in enumerate(layers):

        class DerivedConfig(BaseConfig):
            value = layer['value']
            location = str(i+1)

        configs.append(DerivedConfig)
        keys.append(byoc.Key(DerivedConfig, toggle=layer['toggle']))

    class DummyObj:
        __config__ = configs

        flag = byoc.toggle_param(*keys)

    obj = DummyObj()
    with error:
        assert obj.flag == expected







#!/usr/bin/env python3

import byoc
import pytest
import parametrize_from_file
from param_helpers import *

@parametrize_from_file(
        schema=[
            with_py.eval,
            defaults(kwargs={}),
        ],
)
def test_first_specified(args, kwargs, expected):
    assert byoc.utils.first_specified(*args, **kwargs) == expected

@parametrize_from_file(
        schema=[
            with_py.eval,
            defaults(kwargs={}),
        ],
)
def test_first_specified_err(args, kwargs):
    with pytest.raises(byoc.ApiError) as err:
        byoc.utils.first_specified(*args, **kwargs)

    assert err.match(no_templates)

@parametrize_from_file(schema=with_py.eval)
def test_lookup(x, key, expected):
    assert byoc.utils.lookup(x, key) == expected




#!/usr/bin/env python3

from .model import get_meta

class Meta:

    def __repr__(self):
        loc_str = '' if not self.location else f'location={self.location!r}'
        return f'{self.__class__.__name__}({loc_str})'

    @property
    def location(self):
        pass

class NeverAccessedMeta(Meta):
    pass

class SetAttrMeta(Meta):
    pass

class GetterMeta(Meta):

    def __init__(self, getter):
        self.getter = getter

class LayerMeta(GetterMeta):

    def __init__(self, getter, layer):
        super().__init__(getter)
        self.layer = layer

    @property
    def location(self):
        return self.layer.location

class DefaultMeta(Meta):
    pass

class ExceptionMeta(Meta):

    def __init__(self, exc):
        self.exc = exc

class UnknownMeta(Meta):
    pass



class meta_view:

    def __get__(self, obj, cls=None):
        return _meta_view(obj)

class _meta_view:
    # This class will never refresh the cache, so the results it returns may 
    # change if the value is accessed immediately afterwards.  For this reason, 
    # it is recommended to always access any parameters of interest before 
    # accessing the metadata associated with those parameters.

    def __init__(self, obj):
        self.__obj = obj

    def __getattr__(self, param):
        return get_meta(self.__obj, param)




#!/usr/bin/env python3

import byoc
import pytest
import parametrize_from_file

from byoc.cast import call_with_context
from param_helpers import *

@parametrize_from_file(
        schema=[
            defaults(
                meta='class DummyMeta: pass',
                obj='class DummyObj: pass',
            ),
            cast(
                func=with_byoc.exec(get='f'),
                value=with_py.eval,
                meta=with_py.exec(get=get_meta),
                obj=with_py.exec(get=get_obj),
                expected=with_py.eval,
            ),
            error_or('expected'),
        ],
)
def test_call_with_context(func, value, meta, obj, expected, error):
    context = byoc.Context(value, meta, obj)
    with error:
        assert call_with_context(func, context) == expected

@parametrize_from_file(
        schema=cast(
            obj=with_byoc.exec(get=get_obj, defer=True),
            files=Schema(files.schema)
        ),
        indirect=['files'],
)
def test_relpath(obj, expected, files, monkeypatch):
    monkeypatch.chdir(files)

    obj = obj.exec()
    for param, relpath in expected.items():
        assert getattr(obj, param) == files / relpath

@parametrize_from_file(
        schema=[
            cast(expected=with_py.eval),
            error_or('expected'),
        ],
)
def test_arithmetic_eval(expr, expected, error):
    with error:
        assert byoc.arithmetic_eval(expr) == expected

    with error:
        assert byoc.int_eval(expr) == int(expected)

    with error:
        assert byoc.float_eval(expr) == float(expected)



#!/usr/bin/env python3

import sys
import inspect

from .errors import Error
from more_itertools import first
from pathlib import Path
from typing import Union, Callable, Any

class Context:
    """
    Extra information that can be made available to *cast* functions.

    The *cast* argument to `param` is must be a function that takes one 
    argument and returns one value.  Normally, this argument is simply the 
    value to cast.  However, BYOC will instead provide a `Context` object if 
    the type annotation of that argument is `Context`::

        >>> import byoc
        >>> def f(context: byoc.Context):
        ...     return context.value

    Context objects have the following attributes:

    - :attr:`value`: The value to convert.  This is the same value that would 
      normally be passed directly to the *cast* function.
    - :attr:`meta`: The metadata object associated with the parameter.
    - :attr:`obj`: The object that owns the parameter, i.e. *self*.
    """

    def __init__(self, value, meta, obj):
        self.value = value
        self.meta = meta
        self.obj = obj

def call_with_context(f, context):
    try:
        sig = inspect.signature(f)
        param = first(sig.parameters.values())

    except ValueError:
        pass

    else:
        if param.annotation is Context:
            return f(context)

    return f(context.value)


def relpath(
        context: Context,
        root_from_meta: Callable[[Any], Path]=\
                lambda meta: Path(meta.location).parent,
) -> Path:
    """
    Resolve paths loaded from a file.  Relative paths are interpreted as being 
    relative to the parent directory of the file they were loaded from.

    Arguments:
        context: The context object provided by BYOC to cast functions.
        root_from_meta: A callable that returns the parent directory for 
            relative paths, given a metadata object describing how the value in 
            question was loaded. The default implementation assumes that the 
            metadata object has a :attr:`location` attribute that specifies the 
            path to the relevant file.  This will work if (i) the value was 
            actually loaded from a file and (ii) the default pick function was 
            used (i.e. `first`).  For other pick functions, you may need to 
            modify this argument accordingly.

    Returns:
        An absolute path.
    """
    path = Path(context.value)
    if path.is_absolute():
        return path

    root = root_from_meta(context.meta)
    return root.resolve() / path

def arithmetic_eval(expr: str) -> Union[int, float]:
    """\
    Evaluate the given arithmetic expression.

    Arguments:
        expr:
            The expression to evaluate.  The syntax is identical to python, but 
            only `int` literals, `float` literals, binary operators (except 
            left/right shift, bitwise and/or/xor, and matrix multiplication),
            and unary operators are allowed.

    Returns:
        The value of the given expression.

    Raises:
        SyntaxError: If *expr* cannot be parsed for any reason.
        TypeError: If the *expr* argument is not a string.
        ZeroDivisionError: If *expr* divides by zero.

    It is safe to call this function on untrusted input, as there is no way to 
    construct an expression that will execute arbitrary code.
    """
    import ast, operator

    operators = {
            ast.Add: operator.add,
            ast.Sub: operator.sub,
            ast.Mult: operator.mul,
            ast.Div: operator.truediv,
            ast.FloorDiv: operator.floordiv,
            ast.Mod: operator.mod,
            ast.Pow: operator.pow,

            ast.USub: operator.neg,
            ast.UAdd: operator.pos,
    }

    def eval_node(node):

        if sys.version_info[:2] < (3, 8):
            if isinstance(node, ast.Num):
                return node.n

        else:
            if isinstance(node, ast.Constant):
                if isinstance(node.value, (int, float)):
                    return node.value
                else:
                    err = ArithmeticError(expr, non_number=node.value)
                    err.blame += "{non_number!r} is not a number"
                    raise err

        if isinstance(node, ast.BinOp):
            try:
                op = operators[type(node.op)]
            except KeyError:
                err = ArithmeticError(expr, op=node.op)
                err.blame += "the {op.__class__.__name__} operator is not supported"
                raise err

            left = eval_node(node.left)
            right = eval_node(node.right)
            return op(left, right)

        if isinstance(node, ast.UnaryOp):
            assert type(node.op) in operators
            op = operators[type(node.op)]
            value = eval_node(node.operand)
            return op(value)

        raise ArithmeticError(expr)

    root = ast.parse(expr.lstrip(" \t"), mode='eval')
    return eval_node(root.body)

def int_eval(expr: str) -> int:
    """\
    Same as `arithmetic_eval()`, but convert the result to `int`.
    """
    return int(arithmetic_eval(expr))

def float_eval(expr: str) -> float:
    """\
    Same as `arithmetic_eval()`, but convert the result to `float`.
    """
    return float(arithmetic_eval(expr))

class ArithmeticError(Error, SyntaxError):

    def __init__(self, expr, **kwargs):
        super().__init__(expr=expr, **kwargs)
        self.brief = "unable to evaluate arithmetic expression"
        self.info += "expression: {expr}"





#!/usr/bin/env python3

from . import model
from .model import UNSPECIFIED
from .cast import Context, call_with_context
from .meta import GetterMeta, LayerMeta
from .errors import ApiError, NoValueFound
from operator import attrgetter
from more_itertools import value_chain, always_iterable

class Getter:

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __repr__(self):
        cls = f'byoc.{self.__class__.__name__}'
        args = self.__reprargs__()
        kwargs = [f'{k}={v!r}' for k, v in self.kwargs.items()]
        return f'{cls}({", ".join((*args, *kwargs))})'

    def __reprargs__(self):
        return []  # pragma: no cover

    def bind(self, obj, param):
        raise NotImplementedError

class Key(Getter):

    def __init__(self, config_cls, key=UNSPECIFIED, **kwargs):
        super().__init__(**kwargs)
        self.config_cls = config_cls
        self.key = key

    def __reprargs__(self):
        if self.key is UNSPECIFIED:
            return [self.config_cls.__name__]
        else:
            return [self.config_cls.__name__, repr(self.key)]

    def bind(self, obj, param):
        wrapped_configs = [
                wc for wc in model.get_wrapped_configs(obj)
                if isinstance(wc.config, self.config_cls)
        ]
        return BoundKey(self, obj, param, wrapped_configs)

class ImplicitKey(Getter):

    def __init__(self, wrapped_config, key):
        super().__init__()
        self.key = key
        self.wrapped_config = wrapped_config

    def __reprargs__(self):
        return [repr(self.wrapped_config), repr(self.key)]

    def bind(self, obj, param):
        return BoundKey(self, obj, param, [self.wrapped_config])

class Func(Getter):

    def __init__(self, callable, *, skip=(), dynamic=False, **kwargs):
        super().__init__(**kwargs)
        self.callable = callable
        self.skip = skip
        self.dynamic = dynamic
        self.partial_args = ()
        self.partial_kwargs = {}

    def __reprargs__(self):
        return [repr(self.callable)]

    def partial(self, *args, **kwargs):
        self.partial_args = args
        self.partial_kwargs = kwargs
        return self

    def bind(self, obj, param):
        return BoundCallable(
                self, obj, param,
                self.callable,
                self.partial_args,
                self.partial_kwargs,
                self.dynamic,
                tuple(always_iterable(self.skip)),
        )

class Method(Func):

    def __init__(self, *args, dynamic=True, **kwargs):
        super().__init__(*args, dynamic=dynamic, **kwargs)

    def bind(self, obj, param):
        # Methods used with this getter this will typically attempt to 
        # calculate a value based on other BYOC-managed attributes.  In most 
        # cases, a `NoValueFound` exception will be raised if any of those 
        # attributes is missing a value.  The most sensible thing to do when 
        # this happens is to silently skip this getter, allowing the parameter 
        # that invoked it to continue searching other getters for a value.

        bc = super().bind(obj, param)
        bc.partial_args = (obj, *bc.partial_args)
        bc.exceptions = bc.exceptions or (NoValueFound,)
        return bc

class Attr(Getter):

    def __init__(self, attr, *, skip=(), dynamic=False, **kwargs):
        super().__init__(**kwargs)
        self.attr = attr
        self.skip = skip
        self.dynamic = dynamic

    def __reprargs__(self):
        return [repr(self.attr)]

    def bind(self, obj, param):
        return BoundAttr(
                self, obj, param, self.attr,
                exc=self.skip or (NoValueFound,),
                dynamic=self.dynamic,
        )

class Value(Getter):

    def __init__(self, value, **kwargs):
        super().__init__(**kwargs)
        self.value = value

    def __reprargs__(self):
        return [repr(self.value)]

    def bind(self, obj, param):
        return BoundValue(self, obj, param, self.value)



class BoundGetter:

    def __init__(self, parent, obj, param):
        self.parent = parent
        self.obj = obj
        self.param = param

        # The following attributes are public and may be accessed or modified 
        # by `param` subclasses (e.g. `toggle_param`).  Be careful when making 
        # modifications, though, because any modifications will need to be 
        # re-applied each time the cache expires (because the getters are 
        # re-bound when this happens).
        self.kwargs = parent.kwargs
        self.cast_funcs = list(value_chain(
            self.kwargs.get('cast', []),
            param._get_default_cast()
        ))

        self._check_kwargs()

    def _check_kwargs(self):
        given_kwargs = set(self.kwargs.keys())
        known_kwargs = self.param._get_known_getter_kwargs()
        unknown_kwargs = given_kwargs - known_kwargs

        if unknown_kwargs:
            err = ApiError(
                    getter=self.parent,
                    obj=self.obj,
                    param=self.param,
                    given_kwargs=given_kwargs,
                    known_kwargs=known_kwargs,
                    unknown_kwargs=unknown_kwargs,
            )
            err.brief = f'unexpected keyword argument'
            err.info += lambda e: '\n'.join([
                f"{e.param.__class__.__name__}() allows the following kwargs:",
                *e.known_kwargs,
            ])
            err.blame += lambda e: '\n'.join([
                f"{e.getter!r} has the following unexpected kwargs:",
                *e.unknown_kwargs,
            ])
            raise err

    def iter_values(self, log):
        raise NotImplementedError

    def cast_value(self, value, meta):
        for f in self.cast_funcs:
            context = Context(value, meta, self.obj)
            value = call_with_context(f, context)
        return value

class BoundKey(BoundGetter):

    def __init__(self, parent, obj, param, wrapped_configs):
        super().__init__(parent, obj, param)
        self.key = parent.key
        self.wrapped_configs = wrapped_configs

        if self.key is UNSPECIFIED:
            self.key = param._get_default_key()

    def iter_values(self, log):
        assert self.key is not UNSPECIFIED
        assert self.wrapped_configs is not None

        if not self.wrapped_configs:
            log += f"no configs of class {self.parent.config_cls.__name__}"

        for wrapped_config in self.wrapped_configs:
            config = wrapped_config.config

            if not wrapped_config.is_loaded:
                log += f"skipped {config}: not loaded"
                log += "did you mean to call `byoc.load()`?"
                continue

            if not wrapped_config.layers:
                # If a config has no layers, that probably means an error 
                # occurred when the config was being loaded.  Most likely, the 
                # cause of this error was that some attribute of the object 
                # either wasn't defined, or didn't have an appropriate value.  
                # If the attribute in question was given an appropriate value 
                # after the config was loaded, the config will need to be 
                # reloaded before that value takes effect.
                #
                # That said, I decided to only include a literal description of 
                # the error here, and to leave it to the configs to suggest how 
                # to fix the problem, e.g. by calling `reload()`.  The reason 
                # is that I think a generic message would be wrong/confusing in 
                # too many cases.
                log += f"skipped {config}: loaded, but no layers"
                config.load_status(log)
                continue

            log += f"queried {config}:"
            config.load_status(log)

            for layer in wrapped_config:
                for value in layer.iter_values(self.key, log):
                    yield (
                            value,
                            LayerMeta(self.parent, layer),
                            config.dynamic,
                    )

class BoundCallable(BoundGetter):

    def __init__(self, parent, obj, param, callable, args, kwargs, dynamic, exc=()):
        super().__init__(parent, obj, param)
        self.callable = callable
        self.partial_args = args
        self.partial_kwargs = kwargs
        self.dynamic = dynamic
        self.exceptions = exc

    def iter_values(self, log):
        try:
            value = self.callable(*self.partial_args, **self.partial_kwargs)
        except self.exceptions as err:
            log += f"called: {self.callable}\nraised {err.__class__.__name__}: {err}"
        else:
            log += lambda: f"called: {self.callable}\nreturned: {value!r}"
            yield value, GetterMeta(self.parent), self.dynamic

class BoundAttr(BoundGetter):

    def __init__(self, parent, obj, param, attr, dynamic, exc=()):
        super().__init__(parent, obj, param)
        self.attr = attr
        self.dynamic = dynamic
        self.exceptions = exc

    def iter_values(self, log):
        qualattr = f'{self.obj.__class__.__name__}.{self.attr}'
        try:
            value = getattr(self.obj, self.attr)
        except self.exceptions as err:
            log += f"looked up: {qualattr}\nraised {err.__class__.__name__}: {err}"
        else:
            log += f"looked up: {qualattr}\nreturned: {value!r}"
            yield value, GetterMeta(self.parent), self.dynamic

class BoundValue(BoundGetter):

    def __init__(self, parent, obj, param, value):
        super().__init__(parent, obj, param)
        self.value = value

    def iter_values(self, log):
        log += lambda: f"got hard-coded value: {self.value!r}"
        yield self.value, GetterMeta(self.parent), False









#!/usr/bin/env python3

"""
An object-oriented framework for command-line apps.
"""

__version__ = '0.32.0'

# Define the public API
_pre_import_keys = set()
_pre_import_keys |= set(globals())

from .app import App, BareMeta
from .model import (
        init, load, reload, insert_config, insert_configs, append_config,
        append_configs, prepend_config, prepend_configs, share_configs,
        get_meta,
)
from .params.param import param
from .params.toggle import toggle_param, pick_toggled, Toggle as toggle
from .params.inherited import inherited_param
from .configs.configs import *
from .configs.layers import Layer, DictLayer, FileNotFoundLayer, dict_like
from .configs.attrs import config_attr
from .configs.on_load import on_load
from .getters import Key, Method, Attr, Func, Value
from .pick import first, list, merge_dicts
from .cast import Context, relpath, arithmetic_eval, float_eval, int_eval
from .key import jmes
from .meta import meta_view
from .errors import NoValueFound
from .utils import lookup

# Make everything imported above appear to come from this module:
_post_import_keys = set(globals())
for _key in _post_import_keys - _pre_import_keys:
    globals()[_key].__module__ = 'byoc'
del _pre_import_keys, _post_import_keys, _key

toggle.__name__ = 'toggle'



#!/usr/bin/env python3

import byoc
import pytest
import sys

from parametrize_from_file import Namespace, cast, defaults, error_or
from voluptuous import Schema, And, Or, Optional, Invalid, Coerce
from unittest.mock import Mock
from re_assert import Matches
from pathlib import Path

if sys.version_info[:2] >= (3, 10):
    from functools import partial
    zip_equal = partial(zip, strict=True)
else:
    from more_itertools import zip_equal

with_py = Namespace('from operator import itemgetter', Mock=Mock)
with_byoc = Namespace(byoc, 'from byoc import *', list=list)

class LayerWrapper:

    def __init__(self, layer):
        self.layer = layer

    def __repr__(self):
        return f'LayerWrapper({self.layer!r})'

    def __eq__(self, other):
        def str_or_none(x):
            return str(x) if x is not None else x

        if not isinstance(other, byoc.Layer):
            print('not a layer')
            return False

        if type(self.layer) != type(other):
            print('wrong layer type')
            return False

        if hasattr(self.layer, 'values'):
            if self.layer.values != other.values:
                print('wrong layer values')
                return False

        if hasattr(self.layer, 'location'):
            if str_or_none(self.layer.location) != str_or_none(other.location):
                print('wrong layer location')
                return False

        if hasattr(self.layer, 'root_key'):
            if self.layer.root_key != other.root_key:
                print('wrong layer root key')
                return False

        if hasattr(self.layer, 'schema'):
            if self.layer.schema != other.schema:
                print('wrong layer schema')
                return False

        if hasattr(self.layer, 'linenos'):
            other_linenos = {k: v.line.lineno for k, v in other.linenos.items()}
            if self.layer.linenos != other_linenos:
                print('wrong layer linenos')
                return False

        return True

class DictLayerWrapper(LayerWrapper):

    def __init__(self, *args, **kwargs):
        super().__init__(byoc.DictLayer(*args, **kwargs))

class MetaWrapper:

    def __init__(self, type, location=None, **kwargs):
        self.type = type
        self.location = location
        self.kwargs = kwargs

    def __repr__(self):
        kwargs = {
                'location': self.location,
                **self.kwargs,
        }
        if not self.location:
            del kwargs['location']

        type_str = self.type.__name__
        kwargs_strs = [
                f'{k}={v!r}'
                for k, v in kwargs.items()
        ]

        return f'{self.__class__.__name__}({", ".join([type_str, *kwargs_strs])})'

    def __eq__(self, other):
        if not isinstance(other, self.type):
            print('wrong meta type')
            return False

        if self.location != other.location:
            print('wrong meta location')
            return False

        try:
            getter_cls = self.kwargs['getter']
        except KeyError:
            pass
        else:
            if not isinstance(other.getter, getter_cls):
                print('wrong meta getter')
                return False

        try:
            value = self.kwargs['getter.value']
        except KeyError:
            pass
        else:
            if other.getter.value != value:
                print('wrong meta getter value')
                return False

        try:
            values = self.kwargs['layer.values']
        except KeyError:
            pass
        else:
            if other.layer.values != values:
                print('wrong meta layer values')
                return False

        return True

def eval_obj_layers(layers):
    schema = Schema(empty_ok({
        Coerce(int): eval_config_layers,
    }))
    return schema(layers)

def eval_config_layers(layers):
    schema = Schema(empty_ok([eval_layer]))
    return schema(layers)

def eval_layer(layer):
    schema = Schema(Or(str, {
        'values': with_py.eval,
        'location': str,
        Optional('root_key'): with_py.eval,
        Optional('schema'): with_py.eval,
        Optional('linenos'): with_py.eval,
    }))
    layer = schema(layer)
    layer = with_byoc.eval(layer) if isinstance(layer, str) else \
            byoc.DictLayer(**layer)
    return LayerWrapper(layer)

def eval_meta(meta):
    with_meta = Namespace(
            with_byoc,
            'from byoc.meta import *',
            MetaWrapper=MetaWrapper,
    )
    dict_schema = {
            'type': with_meta.eval,
            Optional('location', default=None): Or(None, str),
            str: with_byoc.eval,
    }
    schema = Schema(Or(dict_schema, with_meta.eval))
    meta = schema(meta)

    if isinstance(meta, MetaWrapper):
        return meta
    elif isinstance(meta, dict):
        return MetaWrapper(**meta)
    else:
        return MetaWrapper(meta)

def collect_layers(obj):
    wrapped_configs = byoc.model.get_wrapped_configs(obj)
    return {
            i: wc.layers
            for i, wc in enumerate(wrapped_configs)
    }

def assert_log_matches(log, expected):
    for log_str, pattern in zip_equal(log.message_strs, expected):
        Matches(pattern).assert_matches(log_str)

def find_param(obj, name=None):
    from more_itertools import only
    class_attrs = obj.__class__.__dict__

    if name:
        return class_attrs[name]
    else:
        params = (
                x for x in class_attrs.values()
                if isinstance(x, byoc.param)
        )
        default = byoc.param()
        default.__set_name__(obj.__class__, '')
        return only(params, default)

def get_obj_or_cls(obj_name, cls_name=None):
    if not cls_name:
        cls_name = f'Dummy{obj_name.title()}'

    def get(ns):
        try:
            return ns[obj_name]
        except KeyError:
            return ns[cls_name]()

    return get

def empty_ok(container):
    return Or(container, And('', lambda y: type(container)()))

get_obj = get_obj_or_cls('obj')
get_config = get_obj_or_cls('config')
get_meta = get_obj_or_cls('meta')
no_templates = '^[^{}]*$'

@pytest.fixture
def files(request, tmp_path):

    # We want to return a `pathlib.Path` instance with an extra `manifest` 
    # attribute that records the specific file names/contents the were 
    # provided.  Unfortunately, `pathlib.Path` uses slots, so the only way to 
    # do this is to make our own subclass.  This is complicated further by the 
    # fact that `pathlib.Path` is just a factory, and instantiates different 
    # classes of objects depending on the operating system.

    class TestPath(type(tmp_path)):
        __slots__ = ('manifest',)

    tmp_path = TestPath._from_parts([tmp_path])
    tmp_path.manifest = request.param

    for name, contents in request.param.items():
        p = tmp_path / name
        p.parent.mkdir(parents=True, exist_ok=True)
        p.write_text(contents)

    return tmp_path

files.schema = empty_ok({str: str})



#!/usr/bin/env python3

import byoc
import pytest
import parametrize_from_file

from param_helpers import *

class DummyObj:
    pass

class DummyConfig(byoc.Config):

    def __init__(self, layers):
        self.layers = layers

    def load(self, obj):
        return layers

@parametrize_from_file(
        schema=[
            cast(
                obj=with_byoc.exec(get=get_obj),
                init_layers=eval_obj_layers,
                load_layers=eval_obj_layers,
                reload_layers=eval_obj_layers,
            ),
            defaults(reload_layers={}),
        ],
)
def test_init_load_reload(obj, init_layers, load_layers, reload_layers):
    if not reload_layers:
        reload_layers = load_layers

    byoc.init(obj)
    assert collect_layers(obj) == init_layers

    try:
        obj.load()
    except AttributeError:
        byoc.load(obj)

    assert collect_layers(obj) == load_layers

    try:
        obj.reload()
    except AttributeError:
        byoc.reload(obj)

    assert collect_layers(obj) == reload_layers

def test_reload_instead_of_init():

    class DummyConfig(byoc.Config):
        def __init__(self, obj):
            super().__init__(obj)
            self.x = 0
        def load(self):
            self.x += 1
            yield byoc.DictLayer({'x': self.x})

    class DummyObj:
        __config__ = [DummyConfig]
        x = byoc.param()

    obj = DummyObj()
    byoc.reload(obj)
    assert obj.x == 1

    byoc.reload(obj)
    assert obj.x == 2

@parametrize_from_file(
        schema=cast(
            obj=with_byoc.exec(get=get_obj),
            layers=eval_obj_layers,
        ),
)
def test_collect_layers(obj, layers):
    assert collect_layers(obj) == layers

def test_share_configs_mutable():

    class DummyConfigA(byoc.Config):
        def load(self):
            yield byoc.DictLayer(self.obj.a, location='a')
    
    class DummyConfigB(byoc.Config):
        def load(self):
            yield byoc.DictLayer(self.obj.b, location='b')

    class DummyDonor:
        __config__ = [DummyConfigB]

    class DummyAcceptor:
        __config__ = [DummyConfigA]

    donor = DummyDonor()
    donor.a = {'x': 1}  # decoy
    donor.b = {'x': 2}

    acceptor = DummyAcceptor()
    acceptor.a = {'x': 3}
    acceptor.b = {'x': 4}  # decoy

    byoc.load(donor)
    byoc.load(acceptor)

    assert collect_layers(donor) == {
            0: [
                DictLayerWrapper({'x': 2}, location='b'),
            ],
    }
    assert collect_layers(acceptor) == {
            0: [
                DictLayerWrapper({'x': 3}, location='a'),
            ],
    }

    byoc.share_configs(donor, acceptor)
    byoc.reload(donor)
    byoc.reload(acceptor)

    assert collect_layers(donor) == {
            0: [
                DictLayerWrapper({'x': 2}, location='b'),
            ],
    }
    assert collect_layers(acceptor) == {
            0: [
                DictLayerWrapper({'x': 3}, location='a'),
            ],
            1: [
                DictLayerWrapper({'x': 2}, location='b'),
            ],
    }

    # Reloading donor updates acceptor:
    donor.b = {'x': 5}
    byoc.reload(donor)

    assert collect_layers(donor) == {
            0: [
                DictLayerWrapper({'x': 5}, location='b'),
            ],
    }
    assert collect_layers(acceptor) == {
            0: [
                DictLayerWrapper({'x': 3}, location='a'),
            ],
            1: [
                DictLayerWrapper({'x': 5}, location='b'),
            ],
    }

def test_get_config_factories():

    sentinel = object()
    class Obj:
        __config__ = sentinel

    obj = Obj()
    assert byoc.model.get_config_factories(obj) is sentinel

@parametrize_from_file(
        schema=cast(
            obj=with_byoc.exec(get=get_obj, defer=True),
            expected=eval_meta,
        ),
)
def test_get_meta(obj, param, expected):
    obj = obj()
    meta = byoc.get_meta(obj, param)
    assert meta == expected



#!/usr/bin/env python3

from .. import model
from ..model import _is_selected_by_cls
from ..errors import NoValueFound, Log
from operator import attrgetter

class config_attr:

    def __init__(self, config_cls=None, *, getter=None):
        self.config_cls = config_cls
        self.getter = getter

    def __set_name__(self, cls, name):
        self.name = name

    def __get__(self, obj, cls=None):
        model.init(obj)

        configs = [x.config for x in model.get_wrapped_configs(obj)]
        getter = self.getter or attrgetter(self.name)

        log = Log()
        log += f"getting '{self.name}' config_attr for {obj!r}"

        for config in configs:
            if not _is_selected_by_cls(config, self.config_cls):
                log += f"skipped {config}: not derived from {self.config_cls.__name__}"
                continue

            try:
                return getter(config)
            except AttributeError as err:
                log += (
                        f"skipped {config}: {getter} raised {err.__class__.__name__}: {err}"
                        if self.getter else
                        f"skipped {config}: {err}"
                )
                continue

        raise NoValueFound("can't find config attribute", log)



#!/usr/bin/env python3

def on_load(f):
    if isinstance(f, type):
        return lambda g: OnLoad(g, f)
    else:
        return OnLoad(f)

class OnLoad:

    def __init__(self, callback, config_cls=None):
        self.callback = callback
        self.config_cls = config_cls

    def __get__(self, obj, cls=None):
        return self.callback.__get__(obj, cls)

    def __call__(self, obj):
        self.callback(obj)

    def is_relevant(self, updated_configs):
        if self.config_cls is None:
            return bool(updated_configs)
        else:
            return any(
                    isinstance(x, self.config_cls)
                    for x in updated_configs
            )







#!/usr/bin/env python3

import pytest, byoc
import parametrize_from_file

from byoc.errors import Log
from byoc.configs.layers import repr_dict_short
from param_helpers import *

values = [
        {'x': 1},
        lambda: {'x': 1},
]
locations = [
        'a',
        lambda: 'a',
]

def test_dict_layer_repr():
    layer = byoc.DictLayer(values={'x': 1}, location='a')
    assert repr(layer) == "DictLayer({'x': 1}, location='a')"

@pytest.mark.parametrize('values', values)
@pytest.mark.parametrize('location', locations)
def test_dict_layer_init(values, location):
    layer = byoc.DictLayer(values=values, location=location)
    assert layer.values == {'x': 1}
    assert layer.location == 'a'

@pytest.mark.parametrize('values', values)
@pytest.mark.parametrize('location', locations)
def test_dict_layer_setters(values, location):
    layer = byoc.DictLayer(values={}, location='')

    layer.values = values
    layer.location = location

    assert layer.values == {'x': 1}
    assert layer.location == 'a'

@parametrize_from_file(
        schema=cast(
            layer=with_byoc.eval(defer=True),
            key=with_py.eval,
            expected=with_py.eval,
        ),
)
def test_layer_iter_values(layer, key, expected, log, monkeypatch):
    monkeypatch.setenv('BYOC_VERBOSE', '1')

    layer = layer()
    actual_log = Log()
    assert list(layer.iter_values(key, actual_log)) == expected
    assert actual_log.message_strs == log

@parametrize_from_file(
        schema=[
            cast(dict=with_py.eval, n_max=int),
            defaults(env={}),
        ],
)
def test_repr_dict_short(dict, env, expected, monkeypatch):
    for k, v in env.items():
        monkeypatch.setenv(k, v)

    assert repr_dict_short(dict) == expected




#!/usr/bin/env python3

import byoc
from byoc import Key, Config, DictLayer, jmes

def test_jmes():

    class DummyConfig(Config):
        def load(self):
            yield DictLayer({'x': {'y': 1}})

    class DummyObj:
        __config__ = [DummyConfig]
        x = byoc.param(
                Key(DummyConfig, jmes('x.y')),
        )

    obj = DummyObj()
    assert obj.x == 1




#!/usr/bin/env python3

from copy import copy
from ..errors import ApiError

class inherited_param:

    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs

    def __set_name__(self, cls, name):
        for super_cls in cls.__mro__[1:]:
            try:
                parent = super_cls.__dict__[name]
                break
            except KeyError:
                pass
        else:
            err = ApiError(
                    cls=cls,
                    name=name,
            )
            err.brief = "no superclass parameter to inherit from"
            err.info += "attempting to create `inherited_param`: {cls.__name__}.{name}"
            err.blame += lambda e: "\n".join((
                    "none of the following exist:", *(
                        f'{x.__qualname__}.{e.name}'
                        for x in e.cls.__mro__[1:]
                    ))
            )
            raise err

        param = copy(parent)
        param._override(self._args, self._kwargs)

        setattr(cls, name, param)




#!/usr/bin/env python3

import byoc
import pytest

def test_app():

    class DummyConfig(byoc.Config):
        def load(self):
            yield byoc.DictLayer({'x': 1})

    class DummyApp(byoc.App):
        __config__ = [DummyConfig]
        x = byoc.param()

        def __bareinit__(self):
            self.y = 0

        def __init__(self, x):
            self.x = x

    app = DummyApp(2)
    assert app.x == 2
    assert app.y == 0

    app = DummyApp.from_bare()
    assert app.x == 1
    assert app.y == 0

def test_app_entrypoint():
    d = {}

    class DummyApp(byoc.App):

        def __bareinit__(self):
            self.x = 1

        def __init__(self, x):
            self.x = x

        def main(self):
            d['x'] = self.x

    DummyApp.entry_point()
    assert d == {'x': 1}

    DummyApp(2).main()
    assert d == {'x': 2}

def test_app_bareinit_defined():

    class DummyApp(byoc.App):

        def __init__(self, x):
            self.x = x

    app = DummyApp.from_bare()
    assert not hasattr(app, 'x')

def test_app_load_reload():
    d = {'x': 1}

    class DummyConfig(byoc.Config):
        autoload = False
        def load(self):
            yield byoc.DictLayer(d)

    class DummyApp(byoc.App):
        __config__ = [DummyConfig]
        x = byoc.param()

    app = DummyApp()

    with pytest.raises(byoc.NoValueFound):
        app.x

    app.load()
    assert app.x == 1

    d['x'] = 2
    assert app.x == 1

    app.reload()
    assert app.x == 2


    



#!/usr/bin/env python3

import byoc
import parametrize_from_file

from byoc.model import UNSPECIFIED
from byoc.errors import Log
from param_helpers import *

class DummyObj:
    pass

class DummyValueIter(byoc.pick.ValuesIter):

    def __init__(self, values):
        self._with_meta = values
        self.log = Log()

    @property
    def with_meta(self):
        return self._with_meta

@parametrize_from_file(
        schema=[
            defaults(
                obj='class DummyObj: pass',
                param='',
                default='',
            ),
            cast(
                getters=Schema(empty_ok([str])),
                expected=Schema(empty_ok([with_py.eval])),
                log=Schema(empty_ok([str])),
            ),
        ],
)
def test_values_iter(obj, param, getters, default, expected, log, monkeypatch):
    monkeypatch.setenv('BYOC_VERBOSE', '1')

    with_obj = with_byoc.exec(obj)
    obj = get_obj(with_obj)
    param = find_param(obj, param)
    getters = [with_obj.eval(x) for x in getters]
    default = with_py.eval(default) if default else UNSPECIFIED

    byoc.init(obj)

    bound_getters = [
            x.bind(obj, param)
            for x in getters
    ]

    values = byoc.pick.ValuesIter(bound_getters, default, Log())
    assert list(values) == expected
    assert values.log.message_strs == log

@parametrize_from_file(
        schema=[
            cast(
                pick_func=with_byoc.eval,
                values_iter=lambda x: DummyValueIter(with_py.eval(x)),
                expected=with_py.eval,
            ),
            with_byoc.error_or('expected'),
        ],
)
def test_pick_functions(pick_func, values_iter, expected, error):
    with error:
        assert pick_func(values_iter) == expected['value']
        assert values_iter.meta == expected['meta']



#!/usr/bin/env python3

import pytest, re
import parametrize_from_file
import byoc

from byoc.errors import Log
from more_itertools import zip_equal, unzip, padded
from param_helpers import *

with_getters = Namespace(
        with_byoc,
        'from byoc.getters import ImplicitKey',
        'from byoc.model import WrappedConfig',
)

@parametrize_from_file(
        schema=cast(getter=with_getters.eval),
)
def test_getter_repr(getter, expected):
    print(repr(getter))
    print(expected)
    assert re.fullmatch(expected, repr(getter))

@parametrize_from_file(
        schema=[
            defaults(
                obj='class DummyObj: pass',
                param='byoc.param()',
                meta='class DummyMeta: pass',
            ),
            cast(
                meta=with_py.exec(get=get_meta),
                given=with_py.eval,
                expected=with_py.eval,
            ),
            with_byoc.error_or('expected'),
        ],
)
def test_getter_cast_value(obj, param, meta, getter, given, expected, error):
    with_obj = with_byoc.exec(obj)
    obj = get_obj(with_obj)
    param = with_obj.eval(param)
    getter = with_obj.eval(getter)

    # Pretend to initialize the descriptor.
    if not hasattr(param, '_name'):
        param.__set_name__(obj.__class__, '')

    byoc.init(obj)
    bound_getter = getter.bind(obj, param)

    with error:
        assert bound_getter.cast_value(given, meta) == expected

@parametrize_from_file(
        schema=[
            defaults(
                obj='class DummyObj: pass',
                param='',
            ),
            cast(
                expected=Schema({
                    'values': with_py.eval,
                    'meta': empty_ok([eval_meta]),
                    'dynamic': empty_ok([with_py.eval]),
                    'log': [str],
                }),
            ),
            error_or('expected'),
        ],
)
def test_getter_iter_values(getter, obj, param, expected, error, monkeypatch):
    monkeypatch.setenv('BYOC_VERBOSE', '1')

    with_obj = with_byoc.exec(obj)
    obj = get_obj(with_obj)
    param = find_param(obj, param)
    getter = with_obj.eval(getter)

    byoc.init(obj)
    bound_getter = getter.bind(obj, param)
    log = Log()

    with error:
        iter = bound_getter.iter_values(log)

        # Can simplify this after more_itertools#591 is resolved.
        try:
            values, metas, dynamic = padded(unzip(iter), [], 3)
        except ValueError:
            values, metas, dynamic = [], [], []
        
        assert list(values) == expected['values']
        assert list(metas) == expected['meta']
        assert list(dynamic) == expected['dynamic']

        assert_log_matches(log, expected['log'])

@parametrize_from_file(
        schema=[
            defaults(
                obj='class DummyObj: pass',
                param='',
            ),
            cast(error=with_byoc.error),
        ],
)
def test_getter_kwargs_err(obj, param, getter, error):
    with_obj = with_byoc.exec(obj)
    obj = get_obj(with_obj)
    param = find_param(obj, param)
    getter = with_obj.eval(getter)

    byoc.init(obj)

    with error:
        getter.bind(obj, param)




#!/usr/bin/env python3

import parametrize_from_file
from param_helpers import *

@parametrize_from_file(
        schema=cast(
            obj=with_byoc.exec(get=get_obj, defer=True),
            expected=with_py.eval,
        ),
)
def test_config_attr(obj, expected):
    obj = obj()
    for attr, value in expected.items():
        assert getattr(obj, attr) == value

@parametrize_from_file(
        schema=cast(
            obj=with_byoc.exec(get=get_obj, defer=True),
            error=with_byoc.error,
        ),
)
def test_config_attr_err(obj, attr, error):
    obj = obj()
    with error:
        getattr(obj, attr)



#!/usr/bin/env python3

import byoc
import pytest
import parametrize_from_file
from param_helpers import *

@parametrize_from_file(
        schema=[
            defaults(error='none', expected={}),
            cast(expected=with_py.eval, error=with_byoc.error),
        ],
)
def test_param(obj, expected, error):
    with error:
        obj = with_byoc.exec(obj, get=get_obj)
        for param, value in expected.items():
            assert getattr(obj, param) == value

@pytest.mark.parametrize('dynamic', [True, False])
def test_param_cache_exc(dynamic):
    # Make sure exceptions are cached just like values are.

    class DummyConfig(byoc.Config):
        def load(self):
            yield byoc.DictLayer(self.obj.values)

    class DummyObj:
        __config__ = [DummyConfig]
        x = byoc.param(dynamic=dynamic)

    obj = DummyObj()

    # Before providing a value:
    obj.values = {}
    assert not hasattr(obj, 'x')

    # After providing a value, before updating the cache:
    obj.values['x'] = 1
    assert (obj.x == 1) if dynamic else not hasattr(obj, 'x')

    # After updating the cache:
    byoc.reload(obj)
    assert obj.x == 1

@pytest.mark.parametrize('dynamic', [True, False])
def test_param_cache_reload(dynamic):

    class BackgroundConfig(byoc.Config):
        def load(self):
            yield byoc.DictLayer(values={'x': -1})

    class ForegroundConfig(byoc.Config):
        values = {'x': 1}

        def load(self):
            # Access the value of the parameter during the load function so 
            # that we can tell if an intermediate value from the loading 
            # process (e.g.  -1) is mistakenly saved as the cache value.
            self.obj.x

            yield byoc.DictLayer(values=self.values)
    
    class DummyObj:
        __config__ = [ForegroundConfig, BackgroundConfig]
        x = byoc.param(dynamic=dynamic)

    
    obj = DummyObj()
    assert obj.x == 1

    # Before updating the cache:
    ForegroundConfig.values['x'] = 2
    assert obj.x == (2 if dynamic else 1)

    # After updating the cache:
    byoc.reload(obj)
    assert obj.x == 2

def test_param_cache_instance_values():
    # Test to make sure the independent instances have independent caches.

    class Background(byoc.Config):
        def load(self):
            yield byoc.DictLayer(values={'x': 1})
    
    class Foreground(byoc.Config):
        autoload = False
        def load(self):
            yield byoc.DictLayer(values={'x': 2})
    
    class DummyObj:
        __config__ = [Foreground, Background]
        x = byoc.param()

    o1 = DummyObj()
    o2 = DummyObj()

    assert o1.x == 1
    assert o2.x == 1

    byoc.model.load(o2)

    assert o1.x == 1
    assert o2.x == 2
    
def test_param_cache_instance_key_map():
    # Test to make sure that key map values, if cached, aren't shared between 
    # instances of different classes.
    
    class DummyConfig(byoc.Config):
        def load(self):
            yield byoc.DictLayer(values={'x': 1})
    
    class DecoyConfig(byoc.Config):
        def load(self):
            yield byoc.DictLayer(values={'x': 2})
    
    class ParentObj:
        x = byoc.param()
    
    class DummyObj(ParentObj):
        __config__ = [DummyConfig]
    
    class DecoyObj(ParentObj):
        __config__ = [DecoyConfig]
    
    decoy = DecoyObj()
    assert decoy.x == 2

    obj = DummyObj()
    assert obj.x == 1

def test_param_cache_get():
    # Test to make sure that the get function is called on every parameter 
    # access, even if the underlying value is cached.
    
    class DummyConfig(byoc.Config):

        def load(self):
            yield byoc.DictLayer(values={'x': 1})
    
    class DummyObj:
        __config__ = [DummyConfig]

        def __init__(self):
            self.y = 0

        def _update_y(self, x):
            self.y += 1
            return x

        x = byoc.param(get=_update_y)
    
    obj = DummyObj()
    assert obj.y == 0

    assert obj.x == 1
    assert obj.y == 1

    assert obj.x == 1
    assert obj.y == 2

    assert obj.x == 1
    assert obj.y == 3

def test_param_set_non_comparable():
    class NonComparable:
        def __eq__(self, other):
            raise AssertionError

    class DummyObj:
        x = byoc.param()

    obj = DummyObj()
    obj.x = nc = NonComparable()

    # This parameter access should not attempt any comparisons.
    assert obj.x is nc


def locals_or_ab(locals):
    if locals:
        shared = dict(byoc=byoc)
        exec(locals, {}, shared)
        return shared

    else:
        class A(byoc.Config): pass
        class B(byoc.Config): pass
        return dict(byoc=byoc, A=A, B=B, a=A(), b=B())

def wrap_key_map(map, x):
    return {
            cls: [
                (key, cast(x))
                for key, cast in keys_casts
            ]
            for cls, keys_casts in map.items()
    }



#!/usr/bin/env python
"""
Shell Doctest module.

:Copyright: (c) 2009, the Shell Doctest Team All rights reserved.
:license: BSD, see LICENSE for more details.
"""

import doctest
import inspect
import re
import subprocess
import sys

master = None
_EXC_WRAPPER = 'shell("%s")'

def shell(cmd):
    p = subprocess.run(
            cmd,
            shell=True,
            universal_newlines=True,  # python3.6-compatible alias for 'text'
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
    )
    print((p.stdout + p.stderr).rstrip())

class ShellExample(doctest.Example):
    def __init__(self, source, want, exc_msg=None, lineno=0, indent=0,
                     label=None,
                     options=None):
        doctest.Example.__init__(self, source, want, exc_msg=None, lineno=lineno, indent=indent,
                     options=None)
        self.label = label

class ShellDocTestParser(doctest.DocTestParser):
    _PROMPT = "$"
    _EXC_WRAPPER = _EXC_WRAPPER
    _EXAMPLE_RE = re.compile(r'''\
        # Source consists of a PS1 line followed by zero or more PS2 lines.
        (?P<source>
            (?:^  (?P<indent> [ ]*))                   # PS0 line: indent
            (?:   \[(?P<label>.+)\]\n)?                # PS0 line: label
            (?:   (?P<user>[\w]*)@(?P<host>[\w\.-]*)\n)? # PS0 line: user@host
            (?:   [ ]* \$ .*)                          # PS1 line
            (?:\n [ ]* \. [ ].*)*)                        # PS2 lines
        \n?
        # Want consists of any non-blank lines that do not start with PS1.
        (?P<want> (?:(?![ ]*$)    # Not a blank line
                     (?![ ]*\$)   # Not a line starting with PS1
                     .*$\n?       # But any other line
                  )*)
        ''', re.MULTILINE | re.VERBOSE)

    def parse(self, string, name='<string>'):
        string = string.expandtabs()
        min_indent = self._min_indent(string)
        if min_indent > 0:
            string = '\n'.join([l[min_indent:] for l in string.split('\n')])
        output = []
        charno, lineno = 0, 0
        for m in self._EXAMPLE_RE.finditer(string):
            output.append(string[charno:m.start()])
            lineno += string.count('\n', charno, m.start())
            (source, options, want, exc_msg) = \
                     self._parse_example(m, name, lineno)
            if not self._IS_BLANK_OR_COMMENT(source):
                source = source.replace("\n","; ")
                user = m.group('user')
                host = m.group('host')
                if host:
                    if user:
                        cmd_base = "ssh %(user)s@%(host)s '%(source)s'"
                    else:
                        cmd_base = "ssh %(host)s '%(source)s'"
                    source = cmd_base % vars()

                example = ShellExample(
                        self._EXC_WRAPPER % source.replace("\n","; "),
                        want,
                        exc_msg,
                        lineno=lineno,
                        label=m.group('label'),
                        indent=min_indent+len(m.group('indent')),
                        options=options,
                )
                output.append(example)

            lineno += string.count('\n', m.start(), m.end())
            charno = m.end()
        output.append(string[charno:])
        return output

    def _parse_example(self, m, name, lineno):
        indent = len(m.group('indent'))
        source_lines = [sl for sl in m.group('source').split('\n') if sl.strip()[1] == " "]
        self._check_prompt_blank(source_lines, indent, name, lineno)
        self._check_prefix(source_lines[1:], ' '*indent + '.', name, lineno)
        source = '\n'.join([sl[indent+len(self._PROMPT)+1:] for sl in source_lines])
        want = m.group('want')
        want_lines = want.split('\n')
        if len(want_lines) > 1 and re.match(r' *$', want_lines[-1]):
            del want_lines[-1]
        self._check_prefix(want_lines, ' '*indent, name,
                           lineno + len(source_lines))
        want = '\n'.join([wl[indent:] for wl in want_lines])
        m = self._EXCEPTION_RE.match(want)
        if m:
            exc_msg = m.group('msg')
        else:
            exc_msg = None
        options = self._find_options(source, name, lineno)
        return source, options, want, exc_msg

    def _check_prompt_blank(self, lines, indent, name, lineno):
        for i, line in enumerate(lines):
            if len(line) >= indent+len(self._PROMPT)+1 and line[indent+len(self._PROMPT)] != ' ':
                raise ValueError('line %r of the docstring for %s '
                                 'lacks blank after %s: %r' %
                                 (lineno+i+1, name,
                                  line[indent:indent+len(self._PROMPT)], line))

if __name__ == "__main__":
    testmod()




#!/usr/bin/env python3

import sys, os, re, inspect, autoprop

from .layers import DictLayer, FileNotFoundLayer, dict_like
from ..utils import first_specified
from ..errors import ApiError
from pathlib import Path
from textwrap import dedent
from more_itertools import one, first
from collections.abc import Iterable

class Config:
    autoload = True
    dynamic = False

    def __init__(self, obj, **kwargs):
        self.obj = obj
        self.autoload = kwargs.pop('autoload', self.autoload)
        self.dynamic = kwargs.pop('dynamic', self.dynamic)
        self.load_status = lambda log: None

        if kwargs:
            raise ApiError(
                    lambda e: f'{e.config.__class__.__name__}() received unexpected keyword argument(s): {", ".join(map(repr, e.kwargs))}',
                    config=self,
                    kwargs=kwargs,
            )

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    @classmethod
    def setup(cls, *args, **kwargs):
        return lambda obj: cls(obj, *args, **kwargs)

    def load(self):
        raise NotImplementedError

class EnvironmentConfig(Config):

    def load(self):
        yield DictLayer(
                values=os.environ,
                location="environment",
        )

class CliConfig(Config):
    autoload = False

@autoprop
class ArgparseConfig(CliConfig):
    parser_getter = lambda obj: obj.get_argparse()
    schema = None

    def __init__(self, obj, **kwargs):
        self.parser_getter = kwargs.pop(
                'parser_getter', unbind_method(self.parser_getter))
        self.schema = kwargs.pop(
                'schema', self.schema)

        super().__init__(obj, **kwargs)

    def load(self):
        args = self.parser.parse_args()
        yield DictLayer(
                values=vars(args),
                schema=self.schema,
                location='command line',
        )

    def get_parser(self):
        # Might make sense to cache the parser.
        return self.parser_getter(self.obj)

    def get_usage(self):
        return self.parser.format_help()

    def get_brief(self):
        return self.parser.description

@autoprop
class DocoptConfig(CliConfig):
    usage_getter = lambda obj: obj.__doc__
    version_getter = lambda obj: getattr(obj, '__version__')
    usage_io_getter = lambda obj: obj.usage_io
    usage_vars_getter = lambda obj: obj.usage_vars
    include_help = True
    include_version = None
    options_first = False
    schema = None

    def __init__(self, obj, **kwargs):
        self.usage_getter = kwargs.pop(
                'usage_getter', unbind_method(self.usage_getter))
        self.version_getter = kwargs.pop(
                'version_getter', unbind_method(self.version_getter))
        self.usage_io_getter = kwargs.pop(
                'usage_io_getter', unbind_method(self.usage_io_getter))
        self.usage_vars_getter = kwargs.pop(
                'usage_vars_getter', unbind_method(self.usage_vars_getter))
        self.include_help = kwargs.pop(
                'include_help', self.include_help)
        self.include_version = kwargs.pop(
                'include_version', self.include_version)
        self.options_first = kwargs.pop(
                'options_first', self.options_first)
        self.schema = kwargs.pop(
                'schema', unbind_method(self.schema))

        super().__init__(obj, **kwargs)

    def load(self):
        import sys, docopt, contextlib

        with contextlib.redirect_stdout(self.usage_io):
            args = docopt.docopt(
                    self.usage,
                    help=self.include_help,
                    version=self.version,
                    options_first=self.options_first,
            )

        # If not specified:
        # - options with arguments will be None.
        # - options without arguments (i.e. flags) will be False.
        # - variable-number positional arguments (i.e. [<x>...]) will be []
        not_specified = [None, False, []]
        args = {k: v for k, v in args.items() if v not in not_specified}

        yield DictLayer(
                values=args,
                schema=self.schema,
                location='command line',
        )

    def get_usage(self):
        from mako.template import Template

        usage = self.usage_getter(self.obj)
        usage = dedent(usage)
        usage = Template(usage, strict_undefined=True).render(
                app=self.obj,
                **self.usage_vars,
        )

        # Trailing whitespace can cause unnecessary line wrapping.
        usage = re.sub(r' *$', '', usage, flags=re.MULTILINE)

        return usage

    def get_usage_io(self):
        try:
            return self.usage_io_getter(self.obj)
        except AttributeError:
            return sys.stdout

    def get_usage_vars(self):
        try:
            return self.usage_vars_getter(self.obj)
        except AttributeError:
            return {}

    def get_brief(self):
        import re
        sections = re.split(
                '\n\n|usage:',
                self.usage,
                flags=re.IGNORECASE,
        )
        return first(sections, '').replace('\n', ' ').strip()

    def get_version(self):
        return self.include_version and self.version_getter(self.obj)


@autoprop
class AppDirsConfig(Config):
    name = None
    config_cls = None
    slug = None
    author = None
    version = None
    schema = None
    root_key = None
    stem = 'conf'

    def __init__(self, obj, **kwargs):
        self.name = kwargs.pop('name', self.name)
        self.config_cls = kwargs.pop('format', self.config_cls)
        self.slug = kwargs.pop('slug', self.slug)
        self.author = kwargs.pop('author', self.author)
        self.version = kwargs.pop('version', self.version)
        self.schema = kwargs.pop('schema', unbind_method(self.schema))
        self.root_key = kwargs.pop('root_key', self.root_key)
        self.stem = kwargs.pop('stem', self.stem)

        super().__init__(obj, **kwargs)

    def load(self):
        for path, config_cls in self.config_map.items():
            yield from config_cls.load_from_path(
                    path=path, schema=self.schema, root_key=self.root_key,
            )

    def get_name_and_config_cls(self):
        if not self.name and not self.config_cls:
            raise ApiError("must specify `AppDirsConfig.name` or `AppDirsConfig.config_cls`")

        if self.name and self.config_cls:
            err = ApiError(
                    name=self.name,
                    format=self.config_cls,
            )
            err.brief = "can't specify `AppDirsConfig.name` and `AppDirsConfig.format`"
            err.info += "name: {name!r}"
            err.info += "format: {format!r}"
            err.hints += "use `AppDirsConfig.stem` to change the filename used by `AppDirsConfig.format`"
            raise err

        if self.name:
            suffix = Path(self.name).suffix
            configs = [
                    x for x in FileConfig.__subclasses__()
                    if suffix in getattr(x, 'suffixes', ())
            ]
            found_these = lambda e: '\n'.join([
                    "found these subclasses:", *(
                        f"{x}: {' '.join(getattr(x, 'suffixes', []))}"
                        for x in e.configs
                    )
            ])
            with ApiError.add_info(
                    found_these,
                    name=self.name,
                    configs=FileConfig.__subclasses__(),
            ):
                config = one(
                        configs,
                        ApiError("can't find FileConfig subclass to load '{name}'"),
                        ApiError("found multiple FileConfig subclass to load '{name}'"),
                )

            return self.name, config

        if self.config_cls:
            return self.stem + self.config_cls.suffixes[0], self.config_cls

    def get_dirs(self):
        from appdirs import AppDirs
        slug = self.slug or self.obj.__class__.__name__.lower()
        return AppDirs(slug, self.author, version=self.version)

    def get_config_map(self):
        dirs = self.dirs
        name, config_cls = self.name_and_config_cls
        return {
                Path(dirs.user_config_dir) / name: config_cls,
                Path(dirs.site_config_dir) / name: config_cls,
        }

    def get_config_paths(self):
        return self.config_map.keys()
        

@autoprop
class FileConfig(Config):
    path = None
    path_getter = lambda obj: obj.path
    schema = None
    root_key = None

    def __init__(self, obj, path=None, *, path_getter=None, schema=None, root_key=None, **kwargs):
        super().__init__(obj, **kwargs)
        self._path = path or self.path
        self._path_getter = path_getter or unbind_method(self.path_getter)
        self.schema = schema or self.schema
        self.root_key = root_key or self.root_key

    def get_paths(self):
        try:
            p = self._path or self._path_getter(self.obj)

        except AttributeError as err:

            def load_status(log, err=err, config=self):
                log += f"failed to get path(s):\nraised {err.__class__.__name__}: {err}"
                if config.paths:
                    br = '\n'
                    log += f"the following path(s) were specified post-load:{br}{br.join(str(p) for p in config.paths)}"
                    log += "to use these path(s), call `byoc.reload()`"

            self.load_status = load_status
            return []


        if isinstance(p, Iterable) and not isinstance(p, str):
            return [Path(pi) for pi in p]
        else:
            return [Path(p)]

    def load(self):
        for path in self.paths:
            yield from self.load_from_path(
                    path=path,
                    schema=self.schema,
                    root_key=self.root_key,
            )

    @classmethod
    def load_from_path(cls, path, *, schema=None, root_key=None):
        try:
            data, linenos = cls._do_load_with_linenos(path)
            yield DictLayer(
                    values=data,
                    linenos=linenos,
                    location=path,
                    schema=schema,
                    root_key=root_key,
            )
        except FileNotFoundError:
            yield FileNotFoundLayer(path)

    @classmethod
    def _do_load_with_linenos(cls, path):
        return cls._do_load(path), {}

    @staticmethod
    def _do_load(path):
        raise NotImplementedError

class YamlConfig(FileConfig):
    suffixes = '.yml', '.yaml'

    @staticmethod
    def _do_load(path):
        import yaml
        with open(path) as f:
            return yaml.safe_load(f)


class TomlConfig(FileConfig):
    suffixes = '.toml',

    @staticmethod
    def _do_load(path):
        import tomli
        with open(path, 'rb') as f:
            return tomli.load(f)


class NtConfig(FileConfig):
    suffixes = '.nt',

    @staticmethod
    def _do_load_with_linenos(path):
        import nestedtext as nt
        keymap = {}
        return nt.load(path, keymap=keymap), keymap

class JsonConfig(FileConfig):
    suffixes = '.json',

    @staticmethod
    def _do_load(path):
        import json
        with open(path) as f:
            return json.load(f)

def unbind_method(f):
    return getattr(f, '__func__', f)



#!/usr/bin/env python3

import functools, os
from inspect import isclass
from ..utils import lookup

class Layer:

    def iter_values(self, key, log):
        raise NotImplementedError

class DictLayer(Layer):

    def __init__(self, values, *, schema=None, root_key=None, location=None, linenos=None):
        # Values:
        # - object that implements `__getitem__()` to either return value 
        #   associated with key, or raise KeyError.
        # - callable that takes no arguments and returns an object matching the 
        #   above description.
        #
        # Location:
        # - string
        # - callable that takes no arguments and returns a string.
        self.values = values
        self.linenos = linenos or {}
        self.schema = schema
        self.root_key = root_key
        self.location = location

    def __repr__(self):
        attrs = {
                'schema': self.schema,
                'root_key': self.root_key,
                'location': self.location,
        }
        attr_strs = [f'{k}={v!r}' for k, v in attrs.items() if v]
        attrs_str = ', '.join([repr(self.values), *attr_strs])
        return f'{self.__class__.__name__}({attrs_str})'

    def iter_values(self, key, log):
        values = self.values

        if self.root_key:
            try:
                values = lookup(values, self.root_key)
            except KeyError:
                log += lambda: format_loc(
                        f"did not find {self.root_key!r} in {repr_dict_short(values)}",
                        self.location,
                )
                return

        if self.schema:
            values = self.schema(values)

        try:
            value = lookup(values, key)
        except KeyError:
            log += lambda: format_loc(
                    f"did not find {key!r} in {repr_dict_short(values)}",
                    self.location,
            )
        else:
            log += lambda: format_loc(
                    f"called: {key!r}\nreturned: {value!r}" if callable(key) else
                    f"found {key!r}: {value!r}",
                    self.location,
            )
            yield value

    @property
    def values(self):
        if self._are_values_deferred:
            self._values = self._values()
            self._are_values_deferred = False
        return self._values

    @values.setter
    def values(self, values):
        self._values = values
        self._are_values_deferred = callable(values)

    @property
    def location(self):
        if self._is_location_deferred:
            self._location = self._location()
            self._is_location_deferred = False
        return self._location

    @location.setter
    def location(self, loc):
        self._location = loc
        self._is_location_deferred = callable(loc)

class FileNotFoundLayer(Layer):

    def __init__(self, path):
        self.location = path

    def iter_values(self, key, log):
        log += f"file does not exist: {self.location}\ndid not find {key!r}"
        return
        yield  # pragma: no cover


def dict_like(*args):

    # I want this function to be usable as a decorator, but I also want to 
    # avoid exposing `__call__()` so that these objects don't look like 
    # deferred values to `Layer`.  These competing requirements necessitate an 
    # awkward layering of wrapper functions and classes.
    
    class dict_like:

        def __init__(self, f, *raises):
            self.f = f
            self.raises = raises

        def __repr__(self):
            return f"{self.__class__.__name__}({self.f!r})"

        def __getitem__(self, key):
            try:
                return self.f(key)
            except tuple(self.raises) as err:
                raise KeyError from err

    is_exception = lambda x: isclass(x) and issubclass(x, Exception)

    if not args:
        return lambda f: dict_like(f)

    elif is_exception(args[0]):
        return lambda f: dict_like(f, *args)

    else:
        return dict_like(*args)

def format_loc(message, loc):
    prefix = f"{loc}:\n" if loc else ""
    return prefix + message

def repr_dict_short(d):
    import sys
    from collections.abc import Mapping
    from textwrap import shorten
    from pprint import pformat

    # Since this function is used for logging, we don't want to crash on 
    # unexpected input.  So we check if the given object is actually a mapping, 
    # and if it's not, we pass it to `pformat()`, which will just use its repr.

    if os.environ.get('BYOC_VERBOSE') or not isinstance(d, Mapping):
        return pformat(d, depth=1, compact=True, width=sys.maxsize)

    try:
        n_max = int(os.environ['BYOC_DICT_KEY_LIMIT'])
    except (KeyError, ValueError):
        n_max = 20

    if len(d) <= n_max:
        key_strs = [f'{k!r}: …' for k in d]
    else:
        n = n_max // 2
        key_strs = [f'{k!r}: …' for k in list(d)[:n]]
        key_strs.append(f'and {len(d) - n} others')

    dict_str = '{' + ', '.join(key_strs) + '}'
    return f'{dict_str}\nTo see the whole dictionary, set the following environment variable: BYOC_VERBOSE=1'



#!/usr/bin/env python3

import byoc
from byoc import Config, DictLayer

class DictConfig(Config):

    def load(self):
        yield DictLayer(values=self.values)

class DictConfigAB(DictConfig):
    values = {'a': 1, 'b': 1}

class DictConfigAC(DictConfig):
    values = {'a': 2, 'c': 2}


def test_easy_1():

    class DummyObj:
        __config__ = [DictConfigAB]
        a = byoc.param()

    obj = DummyObj()
    assert obj.a == 1

def test_easy_2():

    class DummyObj:
        __config__ = [
                DictConfigAB,
                DictConfigAC,
        ]
        a = byoc.param()
        b = byoc.param()
        c = byoc.param()

    obj = DummyObj()
    assert obj.a == 1
    assert obj.b == 1
    assert obj.c == 2
